/*MT*
    
    MediaTomb - http://www.mediatomb.cc/
    
    youtube_video_url.cc - this file is part of MediaTomb.
    
    Copyright (C) 2005 Gena Batyan <bgeradz@mediatomb.cc>,
                       Sergey 'Jin' Bostandzhyan <jin@mediatomb.cc>
    
    Copyright (C) 2006-2010 Gena Batyan <bgeradz@mediatomb.cc>,
                            Sergey 'Jin' Bostandzhyan <jin@mediatomb.cc>,
                            Leonhard Wimmer <leo@mediatomb.cc>
    
    MediaTomb is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License version 2
    as published by the Free Software Foundation.
    
    MediaTomb is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    
    You should have received a copy of the GNU General Public License
    version 2 along with MediaTomb; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.
    
    $Id$
*/

/// \file youtube_video_url.cc
/// \brief Definitions of the Transcoding classes. 

#ifdef HAVE_CONFIG_H
    #include "autoconfig.h"
#endif

#ifdef YOUTUBE

#include <pthread.h>

#include "youtube_video_url.h"
#include "tools.h"
#include "url.h"

using namespace zmm;

#define YOUTUBE_URL_PARAMS_REGEXP   "var swfConfig.*\\;"
#define YOUTUBE_URL_LOCATION_REGEXP "\nLocation: (http://[^\n]+)\n"
#define YOUTUBE_URL_WATCH           "http://www.youtube.com/watch?v="
#define YOUTUBE_URL_PARAM_VIDEO_ID  "video_id"

#define YOUTUBE_URL_EMBED_TAG           "<embed type=\"application/x-shockwave-flash\" +([^>]*)>"
#define YOUTUBE_URL_PARAM_FMT_MAP       ";url_encoded_fmt_stream_map=([^&].*video/x-flv&itag=5)"
#define YOUTUBE_URL_PARAM_FMT_SEPARATE  "([^&].*)&fallback_host.*itag=([0-9]+)"

// url_encoded_fmt_stream_map
#define YOUTUBE_URL_PARAM_FMT         "fmt_url_map"
#define YOUTUBE_IS_HD_AVAILABLE_REGEXP  "IS_HD_AVAILABLE[^:]*: *([^,]*)"
YouTubeVideoURL::YouTubeVideoURL()
{
    curl_handle = curl_easy_init();
    if (!curl_handle)
        throw _Exception(_("failed to initialize curl!\n"));

    EMBED = Ref<RExp>(new RExp());
    EMBED->compile(_(YOUTUBE_URL_EMBED_TAG));
    FMT = Ref<RExp>(new RExp());
    FMT->compile(_(YOUTUBE_URL_PARAM_FMT_MAP));
    FMTURL = Ref<RExp>(new RExp());
    FMTURL->compile(_(YOUTUBE_URL_PARAM_FMT_SEPARATE));
    
    HD = Ref<RExp>(new RExp());
    HD->compile(_(YOUTUBE_IS_HD_AVAILABLE_REGEXP));

    // this is a safeguard to ensure that this class is not called from
    // multiple threads - it is not allowed to use the same curl handle
    // from multiple threads
    pid = pthread_self();
}
    
YouTubeVideoURL::~YouTubeVideoURL()
{
    if (curl_handle)
        curl_easy_cleanup(curl_handle);
}

String YouTubeVideoURL::getVideoURL(String video_id, bool mp4, bool hd)
{
    Ref<StringBuffer> buffer;
    Ref<Matcher> matcher;

    long retcode; 
    String urls;
    String watch;
#ifdef TOMBDEBUG
    bool verbose = true;
#else
    bool verbose = false;
#endif

    if (!string_ok(video_id))
        throw _Exception(_("No video ID specified!"));

    watch = _(YOUTUBE_URL_WATCH) + video_id;

    Ref<URL> url(new URL(YOUTUBE_PAGESIZE));

    buffer = url->download(watch, &retcode, curl_handle, false, verbose, false);
    if (retcode != 200)
        throw _Exception(_("Failed to get URL for video with id ") + watch + _("HTTP response code: ") + String::from(retcode));

    //log_debug("------> REQUEST URL1: %s\n", watch.c_str());
    //log_debug("------> GOT BUFFER1: %s\n", buffer->toString().c_str());

    matcher = EMBED->matcher(buffer->toString());
    if (matcher->next())
    {
        urls = url_unescape(trim_string(matcher->group(1)).c_str());
        urls = url_unescape(urls);
        //log_debug("------> GOT BUFFER EMBED: %s\n", urls.c_str());

        buffer->clear();
        *buffer << urls;
        
        matcher = FMT->matcher(buffer->toString());
        if (matcher->next())
        {
            urls = trim_string(matcher->group(1)).c_str();
            //log_debug("------> GOT BUFFER FMT: %s\n", urls.c_str());
            return whichURL(urls, mp4, hd);
        }
    }
    throw _Exception(_("Could not retrieve YouTube video URL"));
}

String YouTubeVideoURL::whichURL(String tmpStr, bool mp4, bool hd)
{
    while( tmpStr.find("url=") >= 0 )
    {
		Ref<Matcher> matcher;
		String fmturl;
		int length;
		
		tmpStr = tmpStr.substring(tmpStr.find("url=") + 4);
		length = tmpStr.find("url=");
		
		fmturl = (length==-1)? tmpStr : tmpStr.substring(0,length);
		
		//log_debug("------> GOT BUFFER witchURL:(len=%d) %s\n",length, fmturl.c_str() );

	    matcher = FMTURL->matcher( fmturl );
	    if( matcher->next() )
	    {
			int fmt = atoi( trim_string(matcher->group(2)).c_str() );
			
			fmturl = trim_string(matcher->group(1)).c_str();
			
			log_debug("------> GOT BUFFER witchURL: %d -- %s\n", fmt, fmturl.c_str() );
			log_debug("------> PLAYMODE %s %s\n",hd?"SD":"HD",mp4?"flv":"mp4");
			// mp4 22:HD H264+AAC
			if(hd && fmt == 37) return fmturl;
			if(hd && fmt == 45) return fmturl;
			if(hd && fmt == 22) return fmturl;
			// 
			if((hd || mp4) && fmt == 18) return fmturl;
			 // flv 34:H264+AAC low  35:H264+AAC hi
			if(fmt == 5 || fmt == 44 || fmt == 34 || fmt == 35) return fmturl;
		}
		continue;
    }

    throw _Exception(_("Could not retrieve YouTube video URL(wichURL)"));
}

#endif//YOUTUBE
